#include "../emu/h/types.h"
#include "../emu/h/emuPalette.h"
#include "../emu/h/interface.h"
#include "../emu/h/mapper.h"
#include "../emu/h/emuFileHandling.h"
#include "h/ppu.h"

namespace Reminesce
{

char g_convertPPU_Format;
//u8 *myTiles_ = 0;

const unsigned char	AttributeShift[128] =
{
	0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,
	0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,
	4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,
	4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6,4,4,6,6
};

u8 AttributeOffset[32*30];

void AttributeOffsetGenerator()
{
	int i, t;
	for (i=0; i<32*30; i++)
	{
		//	AttributeOffset[i] = ((i & 31) /4) | ((i/16) & 0x38);
		t  = (i & 31) /4;
		t |= (i/16) & 0x38;
		AttributeOffset[i] = t;
	}
}

unsigned char RevBits[256];
void generateRevBits()
{
	for(int i=0; i<256; i++)
	{
		RevBits[i] = 
			(i & 0x80 ? 0x01 : 0) |
			(i & 0x40 ? 0x02 : 0) |
			(i & 0x20 ? 0x04 : 0) |
			(i & 0x10 ? 0x08 : 0) |
			(i & 0x08 ? 0x10 : 0) |
			(i & 0x04 ? 0x20 : 0) |
			(i & 0x02 ? 0x40 : 0) |
			(i & 0x01 ? 0x80 : 0);
	}
}

////////////////////////// Tile Related //////////////////////////

void deleteTiles()
{
/*	if (myTiles_ != 0)
		delete myTiles_;
	myTiles_ = 0;
	*/
}
/*
void convertTiles()
{
	int tile, x, y, t;
	u8 X1, X2, colour;

	deleteTiles();
	if (header.chr == 0)
	{
		g_convertPPU_Format = 0;
		return;
	}

	myTiles_ = new u8[header.chr * 512 * (8*8)];
	
	for (tile = 0; tile < header.chr * 512; tile++) {
	for (y = 0; y<8; y++) 
	{
		X1	= chrromPtr[tile * 16 + y];
		X2	= chrromPtr[tile * 16 + y + 8];
		for (x = 0; x<8; x++) 
		{
			t = 1<<(7 - x);
			colour  = ((X1 & t) != 0);
			colour |= ((X2 & t) != 0)<<1;
			myTiles_[(tile*8 + y)*8 + x] = colour;
		}
	}
	}
	g_convertPPU_Format = 1;
}
*/

/////////////////////////////////////////////////////////////////////////////
template<typename pixelPtr, typename palSize>
void RenderLineAll(u8 scanline, pixelPtr *surface, u16 pitch, palSize thePal);

void RenderTheLine(u8 scanline, void *surface, u16 pitch)
{
	if (pal16)
		RenderLineAll(scanline, (u16*)surface, pitch, pal16);
	else if (pal32)
		RenderLineAll(scanline, (u32*)surface, pitch, pal32);
}

void checkSpriteCollision(u8 scanline)
{
	u16 t;
	u8 tileTable, axisY, axisX, fineY, fineX;
	u8 X1, X2, X3, tile, nt;

	axisY=  sprRam[0];
	tile =  sprRam[1];

	if (ppu.r2000 & 0x20)						//8x16
	{
		if (tile & 1)
			{tileTable = 4; tile--; }
		else
			tileTable = 0;

		if ((scanline-axisY) >7)
			axisY -= 8;							//-8 so you skip the high bits of the tile
	}
	else
		tileTable	= (ppu.r2000 & 8)/2;		//4 if true (2nd tile table)

	tileTable  |= tile/64;
	t = (ppu.r2000 & 0x20) ? 23 : 7;

	if (sprRam[2] & 0x80)						//Horzintal Flip
	{
		X1 = pChrRead[tileTable][((tile & 0x3f) *16) | (u8)(t-(scanline-axisY))];
		X1|= pChrRead[tileTable][((tile & 0x3f) *16) | (u8)(t-(scanline-axisY)+8)];
	}
	else
	{
		X1 = pChrRead[tileTable][((tile & 0x3f) *16) | (u8)(scanline-axisY)];
		X1|= pChrRead[tileTable][((tile & 0x3f) *16) | (u8)(scanline-axisY+8)];
	}

	if (!X1)
		return;

	if (sprRam[2] & 0x40)				//vertical Flip
		X1 = RevBits[X1];

	axisX = sprRam[3];
	if (!(ppu.r2001 & 4))				//if clipped
	{
		if (axisX < 8)
			X1 &= (1<<axisX)-1;
	}
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	nt		= ((ppu.address>>10) & 3) ^ 1;
	fineY	=  ppu.address>>12;
	fineX	=  ppu.xLatch;		//Every scanline fetches it, X fine scroll
	tileTable =  (ppu.r2000 & 16) ? 4 : 0;	//4 is 2nd table (4k each)
	
	t = (axisX+fineX)>>3;
	
	if ((((ppu.address) & 0x1F) + t) > 0x1F)
		tile  = pntBank[nt^1][(ppu.address  & 0x3E0) | ((ppu.address + t) & 0x1F)];
	else
		tile  = pntBank[nt][(ppu.address + t) & 0x3FF];

	t	= tile/64;
	X2	= pChrRead[t + tileTable][((tile & 0x3f) *16) |  fineY ];
	X2 |= pChrRead[t + tileTable][((tile & 0x3f) *16) | (fineY + 8)];

	if ((axisX+fineX) & 0x7)
	{
		X2 <<= ((axisX+fineX) & 0x7);
		
		t = (axisX+fineX)>>3;
		if ((((ppu.address) & 0x1F) + t) > 0x1F)
			tile  = pntBank[nt^1][(ppu.address  & 0x3E0) | ((ppu.address + t) & 0x1F)];
		else
			tile  = pntBank[nt][(ppu.address + t) & 0x3FF];
		t = tile/64;
		//fineY as the pattern bits (X1 & X2)
		X3 = pChrRead[t + tileTable][((tile & 0x3f) *16) |  fineY ];
		X3|= pChrRead[t + tileTable][((tile & 0x3f) *16) | (fineY + 8)];
		X2|= X3>>(8-((axisX+fineX) & 0x7));
	}

	if (!(ppu.r2001 & 2))		//if clipped
	{
		if (axisX < 8)
			X2 &= (1<<axisX)-1;
	}

	if (X1 & X2)
		ppu.sprite0Hit = 0x40;
}

template<typename pixelPtr, typename palSize>
void RenderLineAll(u8 scanline, pixelPtr *surface, u16 pitch, palSize thePal)
{
	u8 sprIndx, sprSize, sprAmount, hasSprLow, SpriteOnLine[8];

	sprIndx = sprAmount = hasSprLow = 0;
	SpriteOnLine[0] = 0xFF;

	if (!(ppu.r2001 & 0x18))
	{
		int x;
		for (x=0; x<256; x++)
			surface[x + scanline*pitch] = 0;
		return;
	}

	if (ppu.r2001 & 0x10)		//sprite are on
	{
		sprSize = (ppu.r2000 & 0x20) ? 16 : 8;
		for (sprIndx=0; sprIndx<64; sprIndx++)
		{
			if ((ui)(scanline - sprRam[sprIndx*4]) < sprSize)
			{
				if (sprAmount<8)
				{
					hasSprLow |= (sprRam[sprIndx*4 + 2] & 0x20);
					SpriteOnLine[sprAmount++]	= sprIndx*4;
				}
				else if (sprAmount == 8)
					{ppu.spriteOver = 0x20; break;}
			}
		}
	}

	if (skipThisFrame == 0)
	{
		if (hasSprLow)
			drawWithLow(scanline, surface, pitch, thePal, sprAmount, SpriteOnLine );
		else
			drawNormally(scanline, surface, pitch, thePal, sprAmount, SpriteOnLine);
	}
	else
		ppu.address ^= 0x400;

	if ((!SpriteOnLine[0]) && (ppu.r2001 & 0x18) && (!ppu.sprite0Hit))
		checkSpriteCollision(scanline);
}

inline void getEmphasisBits(si &emphasis, si &mono)
{
	emphasis = (ppu.r2001 & 0xE0)<<1;
	mono = (ppu.r2001 & 1) ? 0x30 : 0x3F;
}

inline void getTileAndTable(si &sprIndx, si &tile, si &tileTable)
{
	tile =  sprRam[sprIndx + 1];
	if (ppu.r2000 & 0x20)						//8x16
	{
		if (tile & 1)	{tileTable = 4; tile--; }
		else			{tileTable = 0;}
	}
	else
		tileTable	= (ppu.r2000 & 8)/2;		//4 if true (2nd tile table)
}

inline void checkTileHandler(u8 &sprAmount, si &sprIndx, u8 *SpriteOnLine, si &tile, si &tileTable)
{
	int i;
	if (mapInfo->TileHandler)
	{
		for (i = 0; i<sprAmount; i++)
		{
			sprIndx = SpriteOnLine[i];
			{
				getTileAndTable(sprIndx, tile, tileTable);
				mapInfo->TileHandler(tileTable/4, tile, -1);
			}
		}
	}
}

inline void getScrollingVal(si &nt, si &fineY, si &fineX)
{
	nt		= (ppu.address>>10) & 3;
	fineY	=  ppu.address>>12;
	fineX	=  ppu.xLatch;							//Every scanline fetches it, X fine scroll
}

inline void getMiscTileData(si &sprIndx, si &tile, si &tileTable, si &fineX, si &highBits, si &x)
{
	tileTable  |= tile/64;
	fineX		= 0;
	highBits	= (sprRam[sprIndx+2] & 3)<<2;
	x			=  sprRam[sprIndx+3];
}

template<typename pixelPtr, typename palSize>
void drawNormally(u8 scanline, pixelPtr *surface, u16 pitch, palSize *thePal, u8 sprAmount, u8 *SpriteOnLine)
{
	//pitch is actually width
	si i, attributeShift;
	si x, emphasis, mono;
	si tileTable, axisY, fineX, fineY,  colour;
	si X1, X2, t, highBits, tile, tileCount;
	si sprIndx, nt;

	getEmphasisBits(emphasis, mono);
	
	checkTileHandler(sprAmount, sprIndx, SpriteOnLine, tile, tileTable);
	//Back Ground
	if (ppu.r2001 & 8)
	{
		getScrollingVal(nt, fineY, fineX);

		tileCount = 0;
		if (!(ppu.r2001 & 2))							//BG clipping
		{
			tile  = pntBank[nt][ppu.address & 0x3FF];
			if(mapInfo->TileHandler)
			{
				mapInfo->TileHandler((ppu.r2000 & 16) != 0, tile, ppu.address & 0x3FF);
				tileCount++;
			}

			for (x=0; x<8; x++)
			{
				if ((emphasis | (pal[0] & mono)) != 0x3C)
					mono = mono;
				surface[scanline*pitch + x] = thePal[emphasis | (pal[0] & mono)];
			}

			if ((ppu.address & 31)==31)
			{
				nt ^= 1;
				ppu.address ^= 0x41F;
			}
			else
				ppu.address++;
		}
		else
			x = 0;

		tileTable =  (ppu.r2000 & 16) ? 4 : 0;			//4 is 2nd table (4k each)
		while (x<256)
		{
			tile  = pntBank[nt][ppu.address & 0x3FF];

			if(mapInfo->TileHandler)
			{
				mapInfo->TileHandler((ppu.r2000 & 16) != 0, tile, ppu.address & 0x3FF);
				tileCount++;
			}

			attributeShift = AttributeShift[ppu.address & 0x7F];
			t = AttributeOffset[ppu.address & (32*30-1)];

			highBits = (pntBank[nt][960+t] >> attributeShift) & 3;
			highBits = highBits<<2;
			if (g_convertPPU_Format == 0)
			{
				t	= tile/64;
				X1	= pChrRead[t + tileTable][((tile & 0x3f) *16) |  fineY ];
				X2	= pChrRead[t + tileTable][((tile & 0x3f) *16) | (fineY + 8)];
				for (; fineX<8 && x<256; fineX++, x++)
				{
					t = 1<<(7 - fineX);
					colour  = ((X1 & t) != 0) | highBits;
					colour |= ((X2 & t) != 0)<<1;
					//write bytes with colour pass through rgb pal
					//consider it as 16bit
					if (colour & 3)
						surface[scanline*pitch + x] = thePal[emphasis | (pal[colour] & mono)];
					else
						surface[scanline*pitch + x] = thePal[emphasis | (pal[0] & mono)];
				}
			}
			else
			{
				for (; fineX<8 && x<256; fineX++, x++)
				{
					colour  = 
						pMyTile[tileTable + (tile/64)][((tile & 0x3F) * 8 + fineY) * 8 + fineX]
						| highBits;

					//write bytes with colour pass through rgb pal
					//consider it as 16bit
					if (colour & 3)
						surface[scanline*pitch + x] = thePal[emphasis | (pal[colour] & mono)];
					else
						surface[scanline*pitch + x] = thePal[emphasis | (pal[0] & mono)];
				}
			}
			fineX = 0;
			//Increase X and check if on next table
			if ((ppu.address & 31)==31)
			{
				nt ^= 1;
				ppu.address ^= 0x41F;
			}
			else
				ppu.address++;
		}
		if(mapInfo->TileHandler)
		{
			while (tileCount < 33)
			{
				tile  = pntBank[nt][ppu.address & 0x3FF];
				mapInfo->TileHandler((ppu.r2000 & 16) != 0, tile, ppu.address & 0x3FF);
				tileCount++;
			}
		}
	}
	else
	{
		int x;
		for (x=0; x<256; x++)
			surface[scanline*pitch + x] = thePal[emphasis | (pal[0] & mono)];
		ppu.address ^= 0x400;
	}

	//sprites
	for (i = sprAmount-1; i>=0; i--)
	{
		sprIndx = SpriteOnLine[i];
		{
			axisY = sprRam[sprIndx];
			getTileAndTable(sprIndx, tile, tileTable);
			getMiscTileData(sprIndx, tile, tileTable, fineX, highBits, x);

			if (g_convertPPU_Format == 0)
			{
				int lineY;

				if (ppu.r2000 & 0x20)
				{
					if ((scanline-axisY) >7)
						axisY -= 8;							//-8 so you skip the high bits of the tile
					
					lineY = (scanline-axisY);
					if (sprRam[sprIndx+2] & 0x80)			//horzontal Flip
					{
						lineY = 7 - lineY;
						tile ^= 1;
					}
				}
				else
				{
					lineY = (scanline-axisY);
					if (sprRam[sprIndx+2] & 0x80)
						lineY = 7 - lineY;
				}
				

				X1 = pChrRead[tileTable][((tile & 0x3f) *16) + (lineY)];
				X2 = pChrRead[tileTable][((tile & 0x3f) *16) + (lineY+8)];

				if (sprRam[sprIndx+2] & 0x40)				//vertical Flip
					{ X1 = RevBits[X1]; X2 = RevBits[X2];}


				if (!(ppu.r2001 & 4))						//if clipped
				{
					if (x < 8)
					{
						fineX = 8-x;
						x = 8;
					}
				}

				for ( ; (fineX < 8) && (x < 256); x++, fineX++)
				{
 					t = 1<<(7-fineX);
					colour  = ((X1 & t) != 0) | highBits;
					colour |= ((X2 & t) != 0)<<1;
					if (colour & 3)
						surface[x + scanline*pitch] = thePal[emphasis | pal[colour+16] & mono];
				}
			}
			else
			{
				int lineY;

				if (ppu.r2000 & 0x20)
				{
					if ((scanline-axisY) >7)
						axisY -= 8;							//-8 so you skip the high bits of the tile
					
					lineY = (scanline-axisY);
					if (sprRam[sprIndx+2] & 0x80)				//horzontal Flip
					{
						lineY = 7 - lineY;
						tile ^= 1;
					}
				}
				else
				{
					lineY = (scanline-axisY);
					if (sprRam[sprIndx+2] & 0x80)
						lineY = 7 - lineY;
				}
				
				while (lineY >= 16)
				{
					tile++;
					lineY -= 16;
				}
				while (lineY < 0)
				{
					tile--;
					lineY += 16;
				}

				if (!(ppu.r2001 & 4))						//if clipped
				{
					if (x < 8)
					{
						fineX = 8-x;
						x = 8;
					}
				}

				for ( ; (fineX < 8) && (x < 256); x++, fineX++)
				{
					if (sprRam[sprIndx+2] & 0x40)
					{
						colour  = 
							pMyTile[tileTable][((tile & 0x3F) * 8 + lineY) * 8 + 7 - fineX]
							| highBits;
					}
					else
					{
						colour  = 
							pMyTile[tileTable][((tile & 0x3F) * 8 + lineY) * 8 + fineX]
							| highBits;
					}
					if (colour & 3)
						surface[x + scanline*pitch] = thePal[emphasis | pal[colour+16] & mono];
				}
			}
		}
	}
	
	if (ppuClipRightSide && !(ppu.r2001 & 0x2))
	{
		for(int n=248; n<256; n++)
			surface[n + scanline*pitch] = thePal[emphasis | (pal[0] & mono)];
	}
}













template<typename pixelPtr, typename palSize>
void drawWithLow(u8 scanline, pixelPtr *surface, u16 pitch, palSize *thePal, u8 sprAmount, u8 *SpriteOnLine)
{
	si attributeShift;
	si x, emphasis, mono;
	si tileTable, axisY, fineX, fineY,  colour;
	si X1, X2, t, highBits, tile, tileCount;
	si sprIndx, nt;

	si sprSize, indx, pri, tNum;
	u8 sprMask1[33];
	u8 sprMask2[33];
	u8 ntFineX = ppu.xLatch;

	for (x=0; x<33; x++)
		sprMask1[x] = sprMask2[x] = 0;


	getEmphasisBits(emphasis, mono);
	
	for (x=0; x<256; x++)
		#ifdef palType24
				*(long*)((long)surface + scanline*pitch +x*3) = thePal[emphasis | pal[0] & mono];
		#else
				surface[x + scanline*pitch] = thePal[emphasis | pal[0] & mono];
		#endif

	tileCount = 0;
	checkTileHandler(sprAmount, sprIndx, SpriteOnLine, tile, tileTable);


	sprSize = (ppu.r2000 & 0x20) ? 16 : 8;
	for (indx = sprAmount-1; indx>=0; indx--)
	{
		sprIndx = SpriteOnLine[indx];
		{
			axisY=  sprRam[sprIndx];
			getTileAndTable(sprIndx, tile, tileTable);

			if (ppu.r2000 & 0x20)
			{
				if ((scanline-axisY) >7)
					axisY -= 8;							//-8 so you skip the high bits of the tile
			}

			getMiscTileData(sprIndx, tile, tileTable, fineX, highBits, x);
			pri			= (sprRam[sprIndx+2] & 0x20);

			if (sprRam[sprIndx+2] & 0x80)				//vertical Flip
			{
				t  = (ppu.r2000 & 0x20) ? 23 : 7;
				X1 = pChrRead[tileTable][((tile & 0x3f) *16) | (t-(scanline-axisY))];
				X2 = pChrRead[tileTable][((tile & 0x3f) *16) | (t-(scanline-axisY)+8)];
			}
			else
			{
				X1 = pChrRead[tileTable][((tile & 0x3f) *16) | (scanline-axisY)];
				X2 = pChrRead[tileTable][((tile & 0x3f) *16) | (scanline-axisY+8)];
			}

			if (sprRam[sprIndx+2] & 0x40)				//vertical Flip
				{ X1 = RevBits[X1]; X2 = RevBits[X2];}

			t = X1 | X2;
			if (pri)	//low
			{
				sprMask1[(x+ntFineX)/8  ] |= t>>((x + ntFineX) & 7);
				sprMask1[(x+ntFineX)/8+1] |= t<<((8 - ((x + ntFineX) & 7)) & 0xF);
			}
			else
			{
				sprMask2[(x+ntFineX)/8  ] |= t>>((x + ntFineX) & 7);
				sprMask2[(x+ntFineX)/8+1] |= t<<((8 - ((x + ntFineX) & 7)) & 0xF);
			}

			if (!(ppu.r2001 & 4))						//if clipped
			{
				if (x < 8)
				{
					fineX = 8-x;
					x = 8;
				}
			}

			for ( ; (fineX < 8) && (x < 256); x++, fineX++)
			{
 				t = 1<<(7-fineX);
				colour  = ((X1 & t) != 0) | highBits;
				colour |= ((X2 & t) != 0)<<1;
				if (colour & 3)

#ifdef palType24
					*(long*)((long)surface + scanline*pitch + x*3) = (*(long*)((long)surface + scanline*pitch + x*3) & 0xFF000000) | (pal24[emphasis | pal[colour+16] & mono] & 0x00FFFFFF);
#else
					surface[x + scanline*pitch] = thePal[emphasis | pal[colour+16] & mono];
#endif
			}
		}
	}

	for (tNum=0; tNum<33; tNum++)
		sprMask2[tNum] &= 0xFF ^ sprMask1[tNum];

	//Back Ground
	if (ppu.r2001 & 8)
	{
		getScrollingVal(nt, fineY, fineX);
		if (!(ppu.r2001 & 2))							//BG clipping
		{
			x = 8;
			tNum = 1;
			if ((ppu.address & 31)==31)
			{
				nt ^= 1;
				ppu.address ^= 0x41F;
			}
			else
				ppu.address++;
		}
		else
			tNum = x = 0;

		tileTable =  (ppu.r2000 & 16) ? 4 : 0;			//4 is 2nd table (4k each)
		while (x<256)
		{
			tile  = pntBank[nt][ppu.address & 0x3FF];

			if(mapInfo->TileHandler)
			{
				mapInfo->TileHandler((ppu.r2000 & 16) != 0, tile, ppu.address & 0x3FF);
				tileCount++;
			}

			attributeShift = AttributeShift[ppu.address & 0x7F];
			t = AttributeOffset[ppu.address & (32*30-1)];

			highBits = (pntBank[nt][960+t] >> attributeShift) & 3;
			highBits = highBits<<2;

			t	= tile/64;
			X1	= pChrRead[t + tileTable][((tile & 0x3f) *16) |  fineY ];
			X2	= pChrRead[t + tileTable][((tile & 0x3f) *16) | (fineY + 8)];

			X1 &= 0xFF ^ (sprMask2[tNum]);
			X2 &= 0xFF ^ (sprMask2[tNum]);

			tNum++;
			for (; fineX<8 && x<256; fineX++, x++)
			{
				t = 1<<(7 - fineX);
				colour  = ((X1 & t) != 0) | highBits;
				colour |= ((X2 & t) != 0)<<1;
				//write bytes with colour pass through rgb pal
				//consider it as 16bit
				if (colour & 3)
				{
	#ifdef palType24
					*(long*)((long)surface + scanline*pitch + x*3) = (*(long*)((long)surface + scanline*pitch + x*3) & 0xFF000000) | (thePal[emphasis | pal[colour] & mono] & 0x00FFFFFF);
	#else
					surface[scanline*pitch + x] = thePal[emphasis | pal[colour] & mono];
	#endif
				}
			}
			fineX = 0;
			//Increase X and check if on next table
			if ((ppu.address & 31)==31)
			{
				nt ^= 1;
				ppu.address ^= 0x41F;
			}
			else
				ppu.address++;
		}
		if(mapInfo->TileHandler)
		{
			while (tileCount < 33)
			{
				tile  = pntBank[nt][ppu.address & 0x3FF];
				mapInfo->TileHandler((ppu.r2000 & 16) != 0, tile, ppu.address & 0x3FF);
				tileCount++;
			}
		}
	}
	else
	{
		ppu.address ^= 0x400;
	}

	if (ppuClipRightSide && !(ppu.r2001 & 0x2))
	{
		for(int n=248; n<256; n++)
			surface[n + scanline*pitch] = thePal[emphasis | pal[0] & mono];
	}

	if ((!SpriteOnLine[0]) && (ppu.r2001 & 0x18) && (!ppu.sprite0Hit))
		checkSpriteCollision(scanline);
}

}
